/* Copyright (c) 2001-2004 Vaclav Slavik
 */

#include <windows.h>
#include "unicows_import.h"

static HANDLE _fastcall MyLoadLibrary(const char*DllName) {
 UINT em = SetErrorMode(0);
 HANDLE ret = LoadLibraryA(DllName);
 if (!ret) ExitProcess(1);
 SetErrorMode(em);
 return ret;
}

/* ------------------------------------------------------------------------ *
 * Standard implementations:                                                *
 * ------------------------------------------------------------------------ */

static void _declspec(noreturn) __stdcall StdImportError(const char *dll, const char *symbol) {
 char msg[256];
 wsprintfA(msg,"Symbolimport '%s' von DLL '%s'!",symbol,dll);
 MessageBoxA(NULL, msg, NULL, MB_ICONSTOP);
 ExitProcess(1);
}


/* ------------------------------------------------------------------------ *
 * ...and the hooks:                                                        *
 * ------------------------------------------------------------------------ */

void (__stdcall *UnicowsImportError)(const char *, const char *) = StdImportError;


/* ------------------------------------------------------------------------ *
 * GetProcAddress implementation:                                           *
 * ------------------------------------------------------------------------ */

/* NB: unicows.dll implements GetProcAddress, too, but since we need to use
       the *real* GetProcAddress from kernel32.dll, we must load it
       ourselves. And we can't use GetProcAddress for that because it's 
       libunicows stub and not initialized at the time we need it. Therefore
       a hack is done here: we use own reimplementation of GetProcAddress to
       load the symbol from kernel32.dll and use this dynamically loaded
       GetProcAddress from that point on to load all symbols. */

static HMODULE dllHandleKernel32 = 0;

typedef FARPROC (WINAPI *GetProcAddressPtrType)(HINSTANCE,LPCSTR);
static GetProcAddressPtrType ptrGetProcAddress;
    
/* DMC doesn't have UINT_PTR: */
#ifdef __DMC__
    #ifdef _WIN64
        typedef unsigned __int64 UINT_PTR;
    #else
        typedef unsigned int UINT_PTR;
    #endif
#endif

/* GetProcAddress reimplementation
   (code by Alexey A. Popoff <pvax@mail.ru>): */

static FARPROC _fastcall InternalGetProcAddress(HMODULE hModule,
                                      const char *szProcName) {
    /* Helper macro - RVA to pointer */
    #define PTR(baseAdr, RVA) ((UINT_PTR)(baseAdr) + (UINT_PTR)(RVA))

    /* No error checking, no address by ordinal: use only for
       the workaround's needs! */
 unsigned i;
 PIMAGE_DOS_HEADER pDosHeader;
 PIMAGE_NT_HEADERS pNTHeaders;
 PIMAGE_DATA_DIRECTORY pDataDir;
 PIMAGE_EXPORT_DIRECTORY pExportDir;
 WORD* prvaOrdinals;
 DWORD* prvaNames;
 DWORD* prvaFuncs;

 pDosHeader = (PIMAGE_DOS_HEADER)hModule;
 pNTHeaders = (PIMAGE_NT_HEADERS)PTR(hModule, pDosHeader->e_lfanew);

 pDataDir=&(pNTHeaders->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT]);
 pExportDir=(PIMAGE_EXPORT_DIRECTORY)PTR(hModule,pDataDir->VirtualAddress);
 prvaOrdinals = (WORD*)PTR(hModule, pExportDir->AddressOfNameOrdinals);
 prvaNames = (DWORD*)PTR(hModule, pExportDir->AddressOfNames);
 prvaFuncs = (DWORD*)PTR(hModule, pExportDir->AddressOfFunctions);

 for (i = 0; i < pExportDir->NumberOfNames; i++) {
  if (0 == lstrcmpA((LPCSTR)PTR(hModule, prvaNames[i]), szProcName))
    return (FARPROC)PTR(hModule, prvaFuncs[prvaOrdinals[i]]);
 }
 return NULL;
}

static void LoadGetProcAddress(void)
{
    dllHandleKernel32 = MyLoadLibrary("kernel32.dll");
    ptrGetProcAddress = (GetProcAddressPtrType)
        InternalGetProcAddress(dllHandleKernel32, "GetProcAddress");
    if (!ptrGetProcAddress)
        UnicowsImportError("kernel32.dll", "GetProcAddress");
}

/* ------------------------------------------------------------------------ *
 * Implementation of symbols redirection:                                   *
 * ------------------------------------------------------------------------ */

extern void __cdecl _LockUnicowsMutex(void);
extern void __cdecl _UnlockUnicowsMutex(void);

#define DLL_KERNEL32        0
#define DLL_USER32          1
#define DLL_GDI32           2
#define DLL_COMDLG32        3
#define DLL_WINSPOOL        4
#define DLL_SHELL32         5
#define DLL_WINMM           6
#define DLL_ADVAPI32        7
#define DLL_VERSION         8
#define DLL_MPR             9
#define DLL_OLEACC         10
#define DLL_OLEDLG         11
#define DLL_RASAPI32       12
#define DLL_MSVFW32        13
#define DLL_SENSAPI        14
#define DLL_AVICAP32       15
#define DLL_SECUR32        16

#define DLLS_COUNT         17

static HMODULE dllHandles[DLLS_COUNT];
static HMODULE dllUnicowsHandle;
static const char *dllNames[DLLS_COUNT] =
    {
      "kernel32.dll", "user32.dll", "gdi32.dll", "comdlg32.dll",
      "winspool.drv", "shell32.dll", "winmm.dll", "advapi32.dll",
      "version.dll", "mpr.dll", "oleacc.dll", "oledlg.dll", "rasapi32.dll",
      "msvfw32.dll", "sensapi.dll", "avicap32.dll", "secur32.dll"
    };
static int dllsLoaded;
static int useUnicows;

static void LoadDLLs(void)
{
    size_t i;

    LoadGetProcAddress();

    if (GetVersion() >= 0x80000000)
    {
        /* Windows 95/98/ME */
        dllUnicowsHandle = MyLoadLibrary("unicows.dll");
        for (i = 0; i < DLLS_COUNT; i++)
            dllHandles[i] = dllUnicowsHandle;
        useUnicows = 1;
    }

    dllsLoaded = 1;
}

void _stdcall LoadUnicowsSymbol(const char *name, int dll, FARPROC *output) {
 _LockUnicowsMutex();
 if (!dllsLoaded) LoadDLLs();
 if (!dllHandles[dll]) dllHandles[dll] = MyLoadLibrary(dllNames[dll]);
 *output = (*ptrGetProcAddress)(dllHandles[dll], name);

    if (!(*output))
    {
        UnicowsImportError(useUnicows ? "unicows.dll" : dllNames[dll], name);
    }

    _UnlockUnicowsMutex();
}

/* später mal:
	mov	ecx,[esp]
	mov	eax,[ecx+6]
	mov	[output],eax
	movzx	eax,byte ptr[ecx+6+4]
	mov	[dll],eax
	lea	eax,ecx+6+4+1
	mov	[name],eax
*/